diff -Naur ../../linux-2.4.2/linux/Documentation/Configure.help ./Documentation/Configure.help
--- ../../linux-2.4.2/linux/Documentation/Configure.help	Mon Feb 19 10:18:18 2001
+++ ./Documentation/Configure.help	Mon Mar 19 10:36:10 2001
@@ -2328,6 +2328,7 @@
   Say Y here to enable support in the dumb serial driver to support
   the HUB6 card.
 
+
 PCMCIA serial device support
 CONFIG_PCMCIA_SERIAL_CS
   Say Y here to enable support for 16-bit PCMCIA serial devices,
@@ -2354,6 +2355,15 @@
   The module will be called serial_cb.o. If you want to compile it as
   a module, say M here and read Documentation/modules.txt. If unsure,
   say N.
+
+
+Asynchronous I/O support
+CONFIG_AIO
+  Enables support for Posix Asynchronous I/O Facilities. AIO is supported
+  on all filesystems, including those that have enabled Raw I/O
+  (see "Support raw access to SCSI disk devices" under
+  "SCSI support").
+
 
 /dev/agpgart (AGP Support) (EXPERIMENTAL)
 CONFIG_AGP
diff -Naur ../../linux-2.4.2/linux/arch/i386/config.in ./arch/i386/config.in
--- ../../linux-2.4.2/linux/arch/i386/config.in	Mon Jan  8 13:27:56 2001
+++ ./arch/i386/config.in	Mon Mar 19 10:36:10 2001
@@ -171,6 +171,11 @@
 mainmenu_option next_comment
 comment 'General setup'
 
+bool 'Asynchronous I/O support' CONFIG_AIO
+if [ "$CONFIG_AIO" = "y" ]; then
+	int  'System-wide limit on outstanding I/Os' CONFIG_AIO_MAX 4096
+fi
+
 bool 'Networking support' CONFIG_NET
 bool 'SGI Visual Workstation support' CONFIG_VISWS
 if [ "$CONFIG_VISWS" = "y" ]; then
diff -Naur ../../linux-2.4.2/linux/arch/i386/kernel/entry.S ./arch/i386/kernel/entry.S
--- ../../linux-2.4.2/linux/arch/i386/kernel/entry.S	Wed Nov  8 17:09:50 2000
+++ ./arch/i386/kernel/entry.S	Mon Mar 19 10:36:10 2001
@@ -646,6 +646,39 @@
 	.long SYMBOL_NAME(sys_getdents64)	/* 220 */
 	.long SYMBOL_NAME(sys_fcntl64)
 	.long SYMBOL_NAME(sys_ni_syscall)	/* reserved for TUX */
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)	/* 225 */
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)	/* 230 */
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)	/* 235 */
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)	/* 240 */
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)	/* 245 */
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)
+	.long SYMBOL_NAME(sys_ni_syscall)	/* 250 */
+#if defined(CONFIG_AIO)
+	.long SYMBOL_NAME(sys_aio)		/* 251 */
+#else
+	.long SYMBOL_NAME(sys_ni_syscall)
+#endif
 
 	/*
 	 * NOTE!! This doesn't have to be exact - we just have
diff -Naur ../../linux-2.4.2/linux/drivers/scsi/sd.c ./drivers/scsi/sd.c
--- ../../linux-2.4.2/linux/drivers/scsi/sd.c	Sat Feb  3 11:45:55 2001
+++ ./drivers/scsi/sd.c	Mon Mar 19 10:36:10 2001
@@ -36,6 +36,9 @@
 #include <linux/errno.h>
 #include <linux/interrupt.h>
 #include <linux/init.h>
+#if defined(CONFIG_RAW) && defined(CONFIG_AIO)
+#include <linux/aio.h>
+#endif
 
 #include <linux/smp.h>
 
@@ -633,6 +636,31 @@
 	 */
 	scsi_io_completion(SCpnt, good_sectors, block_sectors);
 }
+
+#if defined(CONFIG_RAW) && defined(CONFIG_AIO)
+/*
+ * XXX Interfaces for AIO; ideally done through fops.
+ */
+int
+kaio_is_raw(struct file *filp)
+{
+	return(filp->f_op == &sd_raw_fops);
+}
+
+int
+kaio_raw_wait(void *handle, struct kiobuf *kb)
+{
+	return(sd_raw_wait(handle, kb));
+}
+
+void *
+kaio_raw_rw(int rw, struct file *filp, char *buf, size_t count, loff_t *ppos, 
+	  int *error, struct semaphore *semp, struct kiobuf *iobufp)
+{
+	return(sd_raw_rw(rw, filp, buf, count, ppos, error, semp, iobufp));
+}
+#endif	/* CONFIG_AIO */
+
 /*
  * requeue_sd_request() is the request handler function for the sd driver.
  * Its function in life is to take block device requests, and translate
diff -Naur ../../linux-2.4.2/linux/fs/read_write.c ./fs/read_write.c
--- ../../linux-2.4.2/linux/fs/read_write.c	Fri Feb  9 11:29:44 2001
+++ ./fs/read_write.c	Mon Mar 19 10:36:10 2001
@@ -2,6 +2,12 @@
  *  linux/fs/read_write.c
  *
  *  Copyright (C) 1991, 1992  Linus Torvalds
+ *
+ * KAIO modifications:
+ * 	Copyright 1999, Silicon Graphics, Inc.
+ *
+ * 	Written October 1999 by Rajagopal Ananthanarayanan (ananth)
+ *                      at Silicon Graphics, Inc.
  */
 
 #include <linux/slab.h> 
@@ -13,6 +19,9 @@
 #include <linux/dnotify.h>
 
 #include <asm/uaccess.h>
+#if defined(CONFIG_AIO)
+#include <linux/aio.h>
+#endif
 
 struct file_operations generic_ro_fops = {
 	read:		generic_file_read,
@@ -377,3 +386,87 @@
 bad_file:
 	return ret;
 }
+
+#if defined(CONFIG_AIO)
+
+#ifdef DEBUG
+#define dprintk(x) printk x
+#else
+#define dprintk(x)
+#endif
+
+asmlinkage int sys_aio(int cmd,
+			unsigned long arg1,
+			unsigned long arg2,
+			unsigned long arg3,
+			unsigned long arg4)
+{
+	int ret = 0;
+
+
+	if (!current->kaio_info) {
+		lock_kernel();
+		if (!current->kaio_info && cmd != AIOCMD_SLAVE) {
+			kaio_q_create(current);
+			unlock_kernel();
+			return -ENOTCONN;
+		}
+		unlock_kernel();
+	}
+
+	switch (cmd) {
+	case AIOCMD_READ:
+	case AIOCMD_WRITE:
+	{
+		return(kaio_rw(cmd, (aiocb_t *)arg1, 0));
+	}
+	case AIOCMD_SUSPEND:
+	{
+		/* aio_suspend(aiocb_t *list[], nent, timeout) */
+		return (kaio_suspend((aiocb_t **)arg1, arg2,
+				(struct timespec *)arg3, 1));
+	}
+	case AIOCMD_CANCEL:
+	{
+		/* aio_cancel(int fd, aiocb_t *aiocb); */
+		kaio_key_t kkey;
+
+		if (arg2 && copy_from_user(&kkey, &((aiocb_t *)arg2)->aio_key, sizeof(kaio_key_t))) {
+			ret = -EFAULT;
+			break;
+		}
+		return(kaio_cancel((int)arg1, arg2 ?  &kkey : 0));
+	}
+	case AIOCMD_LIST_IO:
+	{
+		/*
+		 * lio_listio(int mode, aiocb_t *const list[],
+		 * 		int nent, sigevent_t *sig);
+		 */
+		return (kaio_list_io((int)arg1, (struct aiocb **)arg2,
+				   (int)arg3, (sigevent_t *)arg4));
+	}
+	case AIOCMD_PCINVALIDATE:
+	{
+		struct file * file;
+
+		if (!(file = fget(arg1))) {
+			ret = -EINVAL;
+			break;
+		}
+		lock_kernel();
+		truncate_inode_pages(file->f_dentry->d_inode->i_mapping, 0);
+		fput(file);
+		break;
+	}
+	case AIOCMD_SLAVE:
+		lock_kernel();
+		kaio_slave();
+		return 0;
+	default:
+		ret = -EINVAL;
+	}
+	unlock_kernel();
+	return ret;
+}
+#endif /* CONFIG_AIO */
diff -Naur ../../linux-2.4.2/linux/include/asm-i386/errno.h ./include/asm-i386/errno.h
--- ../../linux-2.4.2/linux/include/asm-i386/errno.h	Fri Feb  9 11:40:02 2001
+++ ./include/asm-i386/errno.h	Mon Mar 19 11:07:58 2001
@@ -129,4 +129,8 @@
 #define	ENOMEDIUM	123	/* No medium found */
 #define	EMEDIUMTYPE	124	/* Wrong medium type */
 
+#if defined(CONFIG_AIO)
+#define ECANCELED	125	/* Async I/O: operation canceled */
+#endif /* CONFIG_AIO */
+
 #endif
diff -Naur ../../linux-2.4.2/linux/include/asm-i386/unistd.h ./include/asm-i386/unistd.h
--- ../../linux-2.4.2/linux/include/asm-i386/unistd.h	Fri Aug 11 14:39:23 2000
+++ ./include/asm-i386/unistd.h	Mon Mar 19 11:07:59 2001
@@ -228,6 +228,10 @@
 #define __NR_getdents64		220
 #define __NR_fcntl64		221
 
+#if defined(CONFIG_AIO)
+#define __NR_aio		251
+#endif
+
 /* user-visible error numbers are in the range -1 - -124: see <asm-i386/errno.h> */
 
 #define __syscall_return(type, res) \
diff -Naur ../../linux-2.4.2/linux/include/linux/aio.h ./include/linux/aio.h
--- ../../linux-2.4.2/linux/include/linux/aio.h	Wed Dec 31 16:00:00 1969
+++ ./include/linux/aio.h	Mon Mar 19 11:08:24 2001
@@ -0,0 +1,302 @@
+#ifndef _LINUX_AIO_H
+#define _LINUX_AIO_H
+/*
+ * A kernel implementation of POSIX Asynchronous I/O facility (KAIO).
+ *
+ * Copyright 1999, Silicon Graphics, Inc.
+ *
+ * Written October 1999 by Rajagopal Ananthanarayanan (ananth)
+ *  			at Silicon Graphics, Inc.
+ */
+
+/*
+ * User code gets access to AIO by simply including this header file.
+ * CONFIG_AIO has to be explicitly turned on for the kernel.
+ */
+
+#ifndef __KERNEL__
+#define CONFIG_AIO
+#endif
+
+#if defined(CONFIG_AIO)
+
+#ifdef __KERNEL__
+#include <asm/siginfo.h>
+#else
+#include <sys/types.h> 		/* off_t, etc. 				*/
+#include <signal.h>		/* sigevent_t, etc. 			*/
+#endif
+
+/*
+ * Default & maximum number of threads to be used for processing
+ * parts of aio request. Number of threads can be changed using the
+ * environment variable below.
+ */
+
+#define AIO_THREAD_ENVIRON	"AIO_NTHREADS"
+#define AIO_DEFAULT_THREADS 	 4
+#define AIO_MAX_THREADS		64
+
+#define AIO_TIMES	8
+#define AIO_RESERVED	4
+
+/*
+ * User code can turn on 64-bit offsets in aiocb by:
+ * 	#define _FILE_OFFSET_BITS 64
+ *	#include <features.h>
+ *	#include <linux/aio.h>
+ * which will turn all aio_*() interfaces to operate with 64-bit offsets.
+ * 
+ * Further, #define _LARGEFILE64_SOURCE will enable LFS style aio_*64()
+ * interfaces. Important note: <features.h> should be included as
+ * the first file in a code module, for the above to work correctly.
+ * See <features.h> for further details.
+ *
+ * An alternate way to turn on 64-bit offsets in aiocb is to:
+ *	#define AIO_FILE_OFFSET64
+ *	#include <linux/aio.h>
+ * Similarly LFS interfaces can be enabled by:
+ *	#define AIO_LARGEFILE64
+ *	#include <linux/aio.h>
+ * 
+ * The kernel always uses 64-bit offset values. If user code does not
+ * enable 64-bit offsets, the library will suitably zero out the padding.
+ */
+typedef struct aiocb {
+	int		aio_fildes;	/* File descriptor to perform aio  */ 
+	void		*aio_buf;	/* User's data buffer 		   */
+	size_t		aio_nbytes;	/* No. of bytes in transfer 	   */
+#if defined(__USE_FILE_OFFSET64) || defined(AIO_FILE_OFFSET64) || defined(__KERNEL__)
+	loff_t		aio_offset;
+#else
+	off_t		aio_offset;	/* File position to begin transfer */
+	char		__aio_pad[sizeof(loff_t) - sizeof(off_t)];
+#endif
+	int		aio_reqprio;	/* Unused - Reserved 		   */
+	sigevent_t	aio_sigevent;	/* Notification Options 	   */
+	int		aio_lio_opcode; /* lio_listio opcode 		   */
+	unsigned long	aio_reserved[AIO_RESERVED];/* Don't use this 		   */
+#ifdef KAIO_STATS
+	unsigned long	aio_times[AIO_TIMES];
+#endif
+} aiocb_t;
+
+
+#if defined(__USE_LARGEFILE64) || defined(AIO_LARGEFILE64)
+typedef struct aiocb64 {
+	int		aio_fildes;	/* File descriptor to perform aio  */ 
+	void		*aio_buf;	/* User's data buffer 		   */
+	size_t		aio_nbytes;	/* No. of bytes in transfer 	   */
+	loff_t		aio_offset;	/* File offset - 64 bits	   */
+	int		aio_reqprio;	/* Unused - Reserved 		   */
+	sigevent_t	aio_sigevent;	/* Notification Options 	   */
+	int		aio_lio_opcode; /* lio_listio opcode 		   */
+	unsigned long	aio_reserved[AIO_RESERVED];/* Don't use this 		   */
+#ifdef KAIO_STATS
+	unsigned long	aio_times[AIO_TIMES];
+#endif
+} aiocb64_t;
+#endif
+
+/*
+ * Internal to the aio library
+ */
+#define AIOCMD_READ		1
+#define AIOCMD_SUSPEND		2
+#define AIOCMD_CANCEL		3
+#define AIOCMD_RETURN		4
+#define AIOCMD_PCINVALIDATE	5	/* invalidate page cache	   */
+#define AIOCMD_SLAVE		6	/* put a slave in wait for request */
+#define AIOCMD_WRITE		7
+#define AIOCMD_GENERIC_READ	8 	/* Reserved for internal use	   */
+#define AIOCMD_LIST_IO		9
+
+/*
+ * For gathering timing statistics.
+ */
+#define AIOTIME_REQUEST		0
+#define AIOTIME_IO		1
+#define AIOTIME_COPY		2
+
+#if !defined(__USE_FILE_OFFSET64) && !defined(AIO_FILE_OFFSET64)
+/*
+ * Procedural interfaces.
+ */
+int 	aio_read(struct aiocb *aiocbp);
+int 	aio_write(struct aiocb *aiocbp);
+ssize_t	aio_return(struct aiocb *aiocbp);
+int	aio_cancel(int fildes, struct aiocb *aiocbp);
+int	aio_fsync(int op, struct aiocb *aiocbp);
+int	aio_suspend(const struct aiocb *const list[],
+			 int nent, const struct timespec *timeout);
+int 	lio_listio(int mode, struct aiocb *const list[],
+			int nent, struct sigevent *sig);
+#else
+# ifdef __REDIRECT
+extern int __REDIRECT (aio_read, __P ((struct aiocb *__aiocbp)), aio_read64);
+extern int __REDIRECT (aio_write, __P ((struct aiocb *__aiocbp)), aio_write64);
+extern int __REDIRECT (lio_listio, __P ((int __mode,
+		     struct aiocb *__const __restrict __list[],
+		     int __nent, struct sigevent *__restrict __sig)), lio_listio64);
+extern __ssize_t __REDIRECT (aio_return, __P ((struct aiocb *__aiocbp)), aio_return64);
+extern int __REDIRECT (aio_cancel, __P ((int __fildes, struct aiocb *__aiocbp)),
+		       aio_cancel64);
+extern int __REDIRECT (aio_suspend,
+		       __P ((__const struct aiocb *__const __restrict __list[],
+			     int __nent,
+			     __const struct timespec *__restrict __timeout)),
+		       aio_suspend64);
+#else
+#define aio_read aio_read64
+#define aio_write aio_write64
+#define lio_listio lio_listio64
+#define aio_return aio_return64
+#define aio_cancel aio_cancel64
+#define aio_suspend aio_suspend64
+#endif
+#endif
+
+#define aio_error(a)	(volatile int)(a)->aio_reserved[2]
+/*
+ * 64-bit interfaces.
+ */
+#if defined(__USE_LARGEFILE64) || defined(AIO_LARGEFILE64)
+int 	aio_read64(struct aiocb64 *aiocbp);
+int 	aio_write64(struct aiocb64 *aiocbp);
+ssize_t	aio_return64(struct aiocb64 *aiocbp);
+int	aio_cancel64(int fildes, struct aiocb64 *aiocbp);
+int	aio_fsync64(int op, struct aiocb64 *aiocbp);
+#define aio_error64(a)	aio_error(a)
+int	aio_suspend64(const struct aiocb64 *const list[],
+			 int nent, const struct timespec *timeout);
+int 	lio_listio64(int mode, struct aiocb64 *const list[],
+			int nent, struct sigevent *sig);
+#endif
+
+/*
+ * Return values for aio_cancel.
+ */
+#define AIO_CANCELED		1
+#define AIO_NOTCANCELED		2
+#define AIO_ALLDONE		3
+
+/*
+ * Listio modes
+ */
+#define LIO_WAIT		1
+#define LIO_NOWAIT		2
+
+
+/*
+ * Listio opcodes
+ */
+#define LIO_READ		1
+#define LIO_WRITE		2
+#define LIO_NOP			3
+
+#ifdef __KERNEL__
+
+#if defined(CONFIG_RAW)
+#include <linux/iobuf.h>
+/*
+ * Interfaces expected from Raw I/O for AIO.
+ */
+extern	int kaio_is_raw(struct file *);
+extern	int kaio_raw_wait(void *, struct kiobuf *);
+extern	void *kaio_raw_rw(int, struct file *, char *, size_t, loff_t *, 
+			  int *, struct semaphore *, struct kiobuf *);
+#endif
+
+typedef struct liocb {
+	spinlock_t 	 lio_lock; 	/* protects nent manipulation	   */
+	int	   	 lio_nent; 	/* number of kaio's being awaited  */
+	int	   	 lio_done; 	/* number of kaio's completed	   */
+	wait_queue_head_t lio_wait;	/* wait here for completion of lio */
+	struct sigevent	   lio_sigevent;/* completeion signal info, if any */
+	struct task_struct *lio_task;	/* requesting task		   */
+	int		 lio_mode;	/* mode used in request 	   */
+} liocb_t;
+
+typedef struct kaiocb {
+
+	struct file	*kaio_filp;	/* held by an fget() 		*/
+	void		*kaio_buf;	/* user's data buffer */
+	size_t		kaio_nbytes;
+	aiocb_t		*kaio_uaiocb; /* user aiocb for this request  */
+
+	/* 0x10 */
+	loff_t		kaio_offset;
+	int 		kaio_cmd;	/* read / write 		*/
+	int		kaio_error;
+
+	/* 0x20 */
+	int		kaio_id;	/* unique id for this kaiocb	*/
+	struct kaiocb	*kaio_h_next;	/* hash table pointer		*/
+	struct kaiocb	**kaio_h_pprev;	/* hash table pointer		*/
+	struct page 	**kaio_pages;	/* list of kernel pages   	*/
+
+	/* 0x30 */
+	int		kaio_npages; 	/* no. of kernel pages 	  	*/
+	liocb_t		*kaio_liocb; 	/* List I/O control block pointer */
+	wait_queue_head_t *kaio_suspend_wait;	/* wait here on kaio_suspend */
+	struct page 	*kaio_page_inline; /* one page optimization	*/
+
+	/* 0x40 */
+	struct list_head   kaio_ioq;	/* pointers for queueing aio 	*/
+	struct task_struct *kaio_task;	/* requesting task		*/
+	void		   *kaio_kq;	/* kaioq  of requesting task	*/
+
+	/* 0x50 */
+	sigevent_t	kaio_sigevent;	/* Notification Options		*/
+
+#if defined(CONFIG_RAW)
+	/* RAW I/O fields */
+	struct semaphore kaio_sem;	/* for use by raw I/O		*/
+	struct kiobuf 	 kaio_iobuf; 	/* for use by raw I/O		*/
+	void 		 *kaio_scp;	/* cookie returned by raw I/O   */
+#endif
+
+
+#ifdef KAIO_STATS
+	unsigned long	kaio_times[AIO_TIMES];
+#define kaio_set_time(k, x) 	(k)->kaio_times[(x)] = kaio_time()
+#else
+#define kaio_set_time(k, x)
+#endif /* KAIO_STATS */
+
+} kaiocb_t;
+
+
+typedef struct kaio_key {
+	int	kaio_kid;
+	void	*kaio_kq;
+} kaio_key_t;
+
+
+#define aio_key		aio_reserved[0]
+/* 	aio_key[1]	aio_reserved[1] */
+
+#undef aio_error	/* don't want user definition which is a function */
+#define aio_error	aio_reserved[2]
+#define aio_return	aio_reserved[3]
+
+extern int kaio_rw(int cmd, struct aiocb *, liocb_t *);
+extern int kaio_list_io(int, struct aiocb *[], int, sigevent_t *);
+extern int kaio_suspend(aiocb_t *[], int, struct timespec *, int);
+extern int kaio_cancel(int, kaio_key_t *);
+extern void kaio_init(void);
+extern void kaio_slave(void);
+extern void kaio_q_create(struct task_struct *);
+
+
+/*
+ * Prefix for the names of slave threads
+ */
+#define KAIO_SLAVE_NAME		"kaiod-"
+#define KAIO_SLAVE_NAMELEN	6
+
+#endif /* __KERNEL__ */
+
+#endif /* CONFIG_AIO */
+
+#endif /* _LINUX_AIO_H */
diff -Naur ../../linux-2.4.2/linux/include/linux/sched.h ./include/linux/sched.h
--- ../../linux-2.4.2/linux/include/linux/sched.h	Wed Feb 21 16:09:58 2001
+++ ./include/linux/sched.h	Mon Mar 19 11:08:20 2001
@@ -395,6 +395,11 @@
    	u32 self_exec_id;
 /* Protection of (de-)allocation: mm, files, fs, tty */
 	spinlock_t alloc_lock;
+
+#if defined(CONFIG_AIO)
+	void *kaio_info;
+#endif
+
 };
 
 /*
diff -Naur ../../linux-2.4.2/linux/init/main.c ./init/main.c
--- ../../linux-2.4.2/linux/init/main.c	Wed Jan  3 20:45:26 2001
+++ ./init/main.c	Mon Mar 19 10:36:10 2001
@@ -592,6 +592,13 @@
 #if defined(CONFIG_QUOTA)
 	dquot_init_hash();
 #endif
+#if defined(CONFIG_AIO)
+	{
+		extern void kaio_init(void);
+
+		kaio_init();
+	}
+#endif
 	check_bugs();
 	printk("POSIX conformance testing by UNIFIX\n");
 
diff -Naur ../../linux-2.4.2/linux/kernel/exit.c ./kernel/exit.c
--- ../../linux-2.4.2/linux/kernel/exit.c	Fri Feb  9 11:29:44 2001
+++ ./kernel/exit.c	Mon Mar 19 10:36:10 2001
@@ -436,6 +436,13 @@
 #ifdef CONFIG_BSD_PROCESS_ACCT
 	acct_process(code);
 #endif
+#if defined(CONFIG_AIO)
+	if (tsk->kaio_info) {
+		extern void kaio_exit(void);
+
+		kaio_exit();
+	}
+#endif
 	__exit_mm(tsk);
 
 	lock_kernel();
diff -Naur ../../linux-2.4.2/linux/kernel/fork.c ./kernel/fork.c
--- ../../linux-2.4.2/linux/kernel/fork.c	Fri Feb  9 11:29:44 2001
+++ ./kernel/fork.c	Mon Mar 19 10:36:10 2001
@@ -676,6 +676,13 @@
 	if (!current->counter)
 		current->need_resched = 1;
 
+#if defined(CONFIG_AIO)
+	if (current->kaio_info) {
+		extern int kaio_fork(struct task_struct *, int);
+
+		kaio_fork(p, clone_flags);
+	}
+#endif
 	/*
 	 * Ok, add it to the run-queues and make it
 	 * visible to the rest of the system.
diff -Naur ../../linux-2.4.2/linux/mm/filemap.c ./mm/filemap.c
--- ../../linux-2.4.2/linux/mm/filemap.c	Fri Feb 16 16:06:17 2001
+++ ./mm/filemap.c	Mon Mar 19 11:03:35 2001
@@ -2,6 +2,13 @@
  *	linux/mm/filemap.c
  *
  * Copyright (C) 1994-1999  Linus Torvalds
+ * 
+ * KAIO modifications:
+ * 	Copyright 1999, Silicon Graphics, Inc.
+ *
+ * 	Written October 1999 by Rajagopal Ananthanarayanan (ananth)
+ *                      at Silicon Graphics, Inc.
+ * 	With modifications suggested by Doug Rady at Silicon Graphics, Inc.
  */
 
 /*
@@ -28,6 +35,25 @@
 
 #include <linux/highmem.h>
 
+#if defined(CONFIG_AIO)
+#include <linux/aio.h>
+
+#ifdef AIO_DEBUG
+#define	assert(expression) { \
+	if (!(expression)) { \
+		(void)printk(KERN_ERR \
+			"assertion \"%s\" failed: file \"%s\", line %d\n", \
+			#expression, \
+			__FILE__, __LINE__); \
+	} \
+}
+#else
+#define assert(expresssion)
+#endif
+
+#define pageof(x)	(x >> PAGE_SHIFT)
+#endif /* CONFIG_AIO */
+
 /*
  * Shared mappings implemented 30.11.1994. It's not fully working yet,
  * though.
@@ -1229,6 +1255,1207 @@
 	desc->buf += size;
 	return size;
 }
+
+#if defined(CONFIG_AIO)
+
+static unsigned long 	kaio_hash_mask = 0;
+static kaiocb_t 	**kaio_hash_table;
+static int		kaio_hash_table_size;
+static int		kaio_number = 1;  	/* id for kaio 		*/
+static int		kaio_nr_hashed = 0;
+static spinlock_t 	kaio_hash_lock = SPIN_LOCK_UNLOCKED;
+
+typedef struct {
+	spinlock_t kaioq_lock;
+	struct semaphore kaiosem;
+
+	/* 0x10 */
+	struct list_head kaioq;
+	int		 kaioqref;
+	volatile int	 kaioq_nslaves;
+
+	/* 0x20 */
+	struct semaphore kaioq_exit_sem;
+} kaio_queue_t;
+
+#define KAIO_INFO(task) ((kaio_queue_t *) (task)->kaio_info)
+
+/*
+ * AIO HASH TABLE:
+ * 	A simple system wide table hashed on kq & aio_id. 
+ *	Most of this code adapted from the buffer cache hash table.
+ *	All operations protected by kernel lock.
+ */
+#define _kaio_hashfn(kq, kid)	((unsigned)(((unsigned)kq)^(kid)) & kaio_hash_mask)
+#define kaio_hash(kq, kid)	kaio_hash_table[_kaio_hashfn((kq),(kid))]
+
+static void
+kaio_init_hash_table(void)
+{
+	int order = 1;		/* Currently maximum order.. */
+
+	kaio_hash_table_size = (1UL << order) * PAGE_SIZE / sizeof(kaiocb_t *);
+	kaio_hash_table = (kaiocb_t  **) __get_free_pages(GFP_ATOMIC, order);
+	
+	if (!kaio_hash_table)
+		panic("Failed to allocate kaio hash table\n");
+	memset(kaio_hash_table, 0, kaio_hash_table_size * sizeof(kaiocb_t *));
+	kaio_hash_mask = kaio_hash_table_size-1;
+}
+
+static void
+kaio_add_hash(kaiocb_t *kaiocb)
+{
+	kaiocb_t **khp;
+	kaiocb_t *next;
+
+	spin_lock(&kaio_hash_lock);
+	kaiocb->kaio_id = kaio_number++; /* XXX: how about overflow? */
+	khp = &kaio_hash(current->kaio_info, kaiocb->kaio_id);
+	next = *khp;
+
+	kaiocb->kaio_h_next = NULL;
+	kaiocb->kaio_h_pprev = NULL;
+
+	if (next) {
+		kaiocb->kaio_h_next = next;
+		next->kaio_h_pprev = &kaiocb->kaio_h_next;
+	}
+	*khp = kaiocb;
+	kaiocb->kaio_h_pprev = khp;
+	kaio_nr_hashed++;
+	spin_unlock(&kaio_hash_lock);
+}
+
+static kaiocb_t *
+kaio_find_hash(kaio_key_t *kkey)
+{		
+	kaiocb_t *next;
+
+	next = kaio_hash(KAIO_INFO(current), kkey->kaio_kid);
+	for (;;) {
+		kaiocb_t *tmp = next;
+		if (!next)
+			break;
+		next = tmp->kaio_h_next;
+		if (tmp->kaio_id != kkey->kaio_kid)
+			continue;
+		assert(tmp->kaio_task->kaio_info == kkey->kaio_kq);
+		next = tmp;
+		break;
+	}
+	return next;
+}
+
+static inline void
+kaio_remove_hash(kaiocb_t *kaiocb)
+{
+	kaiocb_t **pprev = kaiocb->kaio_h_pprev;
+	if (pprev) {
+		kaiocb_t *next = kaiocb->kaio_h_next;
+		if (next) {
+			next->kaio_h_pprev = pprev;
+			kaiocb->kaio_h_next = NULL;
+		}
+		*pprev = next;
+		kaiocb->kaio_h_pprev = NULL;
+	}
+	kaio_nr_hashed--;
+}
+
+/*
+ * PER-TASK AIO QUEUE OPERATIONS.
+ * 	Proceted by a spinlock embedded in the queue.
+ */
+
+
+/*
+ * Create: an aio queue for the given task. A q create is attempted as 
+ * part of all aio operations, including the slave start operation.
+ * The queue is freed as part of exit. 
+ */
+void
+kaio_q_create(struct task_struct *task)
+{
+	struct task_struct 	*p;
+	kaio_queue_t		*kq;
+
+	/* protected by lock_kernel */
+	if (KAIO_INFO(task))
+		return;
+
+	kq = kmalloc(sizeof(kaio_queue_t), GFP_KERNEL);
+
+	if (!kq) {
+		printk("kmalloc KAIO_INFO(task) failed!!\n");
+		do_exit(SIGKILL);
+	}
+	/*
+	 * XXX: the following memset is required
+	 * in spite of individual member initializaion of
+	 * kaio_queue_t. why?
+	 */
+	memset(&kq->kaioq_lock, 0, sizeof(kaio_queue_t));
+	sema_init(&kq->kaiosem, 0);
+	sema_init(&kq->kaioq_exit_sem, 0);
+	INIT_LIST_HEAD(&kq->kaioq);
+	spin_lock_init(&kq->kaioq_lock);
+
+	kq->kaioq_nslaves = -1; /* -1 indicates no slaves yet */
+	/*
+	 * If there are any clones, propage kaio_info.
+	 * Following will automatically set current task kaioq & reference.
+	 */
+	kq->kaioqref = 0;
+	for_each_task(p) {
+		if (p->mm == current->mm) {
+			KAIO_INFO(p) = kq;
+			kq->kaioqref++;
+		}
+	}
+	
+}
+
+
+/*
+ * Enqueue: now the aio is officially enqueued for I/O, accessible to the slave
+ * 	threads, and is also entered into the hash table, accessible to
+ * 	user operations such as aio_suspend.
+ */
+static int
+kaio_enqueue(kaio_queue_t *kq, kaiocb_t *kaiocb)
+{
+	kaiocb->kaio_error = EINPROGRESS;
+	if (put_user(kaiocb->kaio_error, &kaiocb->kaio_uaiocb->aio_error))
+		return(-EFAULT);
+	kaiocb->kaio_suspend_wait = 0;
+	kaio_set_time(kaiocb, AIOTIME_REQUEST);
+	INIT_LIST_HEAD(&kaiocb->kaio_ioq);
+
+	spin_lock(&kq->kaioq_lock);
+	if (kq->kaioq_nslaves == 0) {
+		spin_unlock(&kq->kaioq_lock);
+		return -EAGAIN;
+		
+	}
+	kaio_add_hash(kaiocb);
+	list_add(&kaiocb->kaio_ioq, &kq->kaioq);
+	up(&kq->kaiosem);
+	spin_unlock(&kq->kaioq_lock);
+
+	return 0;
+}
+
+/*
+ * Wait for an aio and dequeue. Normally, kaiosem counts the number
+ * of entries available in the queue; however, kaio_cancel can
+ * dequeue entries as well, and it does not down the semaphore,
+ * because there is no easy way to synchronize this action with kaio_dequeue.
+ * Instead, the dequeue operation simply retries if it finds the queue
+ * empty eventhough the semaphore count was non-zero (i.e. the down succeded).
+ * The down() is interrupted when the parent sends a signal when it exits.
+ */
+static kaiocb_t *
+kaio_dequeue(kaio_queue_t *kq)
+{
+	struct list_head *tmp;
+
+retry:
+	if (down_interruptible(&kq->kaiosem))
+		return 0;
+
+	spin_lock(&kq->kaioq_lock);
+	tmp = kq->kaioq.prev;
+	if (tmp != &kq->kaioq) {
+		list_del(tmp);
+		spin_unlock(&kq->kaioq_lock);
+		INIT_LIST_HEAD(tmp);
+		return(list_entry(tmp, struct kaiocb, kaio_ioq));
+	} else {
+		spin_unlock(&kq->kaioq_lock);
+		goto retry;
+	}
+}
+
+static int
+kaio_dequeue_kaiocb(kaio_queue_t *kq, kaiocb_t *kaiocb)
+{
+	spin_lock(&kq->kaioq_lock);
+	if (kaiocb->kaio_ioq.prev == &kaiocb->kaio_ioq) {
+		assert(kaiocb->kaio_ioq.next == &kaiocb->kaio_ioq);
+		spin_unlock(&kq->kaioq_lock);
+		return 0;
+	}
+	list_del(&kaiocb->kaio_ioq);
+	INIT_LIST_HEAD(&kaiocb->kaio_ioq);
+	spin_unlock(&kq->kaioq_lock);
+	return 1;
+}
+
+static kaiocb_t *
+kaio_dequeue_fd(kaio_queue_t *kq, struct file *filp)
+{
+	struct list_head *tmp;
+	kaiocb_t *kaiocb;
+
+	spin_lock(&kq->kaioq_lock);
+	tmp = kq->kaioq.next; /* start from first of list; why not? */
+	while (tmp != &kq->kaioq) {
+		kaiocb = list_entry(tmp, struct kaiocb, kaio_ioq);
+		if (kaiocb->kaio_filp == filp) {
+			list_del(tmp);
+			spin_unlock(&kq->kaioq_lock);
+			INIT_LIST_HEAD(tmp);
+			return(kaiocb);
+		}
+		tmp = tmp->next;
+	}
+	spin_unlock(&kq->kaioq_lock);
+	return 0;
+}
+
+/*
+ * Boot time initialization.
+ */
+void
+kaio_init(void)
+{
+	kaio_init_hash_table();
+}
+
+static inline unsigned long
+kaio_time(void)
+{
+	struct timeval tv;
+
+	do_gettimeofday(&tv);
+	return (tv.tv_sec * 1000000 + tv.tv_usec);
+}
+
+static void
+send_signal(sigevent_t *sigev, struct task_struct *task)
+{
+	if (sigev->sigev_notify == SIGEV_SIGNAL) {
+		struct siginfo sinfo;
+		
+		memset(&sinfo, 0, sizeof(sinfo));
+		sinfo.si_signo = sigev->sigev_signo;
+		sinfo.si_code = SI_ASYNCIO;
+		sinfo.si_value = sigev->sigev_value;
+
+		send_sig_info(sigev->sigev_signo, &sinfo, task);
+	}
+}
+
+static int
+lio_complete(liocb_t *liocb, int n, int wakeup)
+{
+	spin_lock(&liocb->lio_lock);
+	liocb->lio_done += n;
+	if (liocb->lio_nent == liocb->lio_done) {
+		spin_unlock(&liocb->lio_lock);
+		if (wakeup) {
+			lock_kernel();
+			wake_up(&liocb->lio_wait);
+			unlock_kernel();
+		}
+		send_signal(&liocb->lio_sigevent, liocb->lio_task);
+		if (liocb->lio_mode == LIO_NOWAIT)
+			kfree(liocb);
+		return 1;
+	}
+	if (!wakeup)
+		lock_kernel();
+	spin_unlock(&liocb->lio_lock);
+	return 0;
+}
+
+static int
+kaio_io_complete(kaiocb_t *kaiocb, int nbytes, int ret, int gc)
+{
+	aiocb_t	*uaiocb = kaiocb->kaio_uaiocb;
+	liocb_t	*liocb = kaiocb->kaio_liocb;
+	int	i;
+
+	if (kaiocb->kaio_pages) {
+
+		for (i = 0; i < kaiocb->kaio_npages; i++)
+			page_cache_release(kaiocb->kaio_pages[i]);
+		if (kaiocb->kaio_pages != &kaiocb->kaio_page_inline)
+			kfree(kaiocb->kaio_pages);
+	}
+
+#ifdef KAIO_STATS
+	if (copy_to_user(&uaiocb->aio_times, &kaiocb->kaio_times,
+			sizeof(unsigned long) * AIO_TIMES))
+	{
+		ret = -EFAULT;
+	} else
+#endif
+	if (put_user(nbytes, &uaiocb->aio_return))
+		ret = -EFAULT;
+	/* aio_error does not go through syscall so -ret */
+	else if (put_user(-ret, &uaiocb->aio_error))
+		ret = -EFAULT;
+	else {
+		send_signal(&kaiocb->kaio_sigevent, kaiocb->kaio_task);
+		spin_lock(&kaio_hash_lock);
+		if (kaiocb->kaio_suspend_wait) {
+			lock_kernel();
+			kaiocb->kaio_error = -ret;
+			wake_up(kaiocb->kaio_suspend_wait);
+			unlock_kernel();
+		} else {
+			kaiocb->kaio_error = -ret;
+			/* gc |= (liocb && liocb->lio_mode == LIO_WAIT); */
+			gc = 1;
+			if (gc)
+				kaio_remove_hash(kaiocb);
+		}
+		spin_unlock(&kaio_hash_lock);
+		if (liocb)
+			lio_complete(liocb, 1, 1);
+	}
+	if (gc) {
+		fput(kaiocb->kaio_filp);
+		kfree(kaiocb);
+	}
+	return ret;
+}
+
+static int
+kaio_slave_suspend(kaiocb_t *kaiocb)
+{
+	off_t pos;
+	int i, ret = 0, nbytes, isize;
+	struct page *page;
+
+	nbytes = kaiocb->kaio_nbytes;
+	isize = kaiocb->kaio_filp->f_dentry->d_inode->i_size;
+	for (i = 0; i < kaiocb->kaio_npages; i++) {
+		page = kaiocb->kaio_pages[i];
+		wait_on_page(page);
+		/*
+		 * Ok, we have the page. 
+		 */
+		if (!Page_Uptodate(page) || PageError(page)) {
+			ret = -EIO; /* Some unspecfied error */
+			goto error;
+		}
+	}
+	kaio_set_time(kaiocb, AIOTIME_IO);
+
+	pos = kaiocb->kaio_offset;
+	i = 0;
+	page = kaiocb->kaio_pages[0];
+	assert(nbytes);
+
+	while  (nbytes && pos < isize) {
+		unsigned long offset, nr, left;
+
+		offset = pos & ~PAGE_MASK;
+		nr = PAGE_SIZE - offset;
+		if (nr > isize - pos)
+			nr = isize - pos;
+		if (nbytes < nr)
+			nr = nbytes;
+
+		left = __copy_to_user(kaiocb->kaio_buf,
+				(char *) page_address(page) + offset, nr);
+		nbytes -= (nr - left);
+		if (left) {
+			ret = -EFAULT;
+			goto error; 
+		}
+		pos += nr;
+		kaiocb->kaio_buf = (char *)kaiocb->kaio_buf + nr;
+		if ((pos & ~PAGE_MASK) == 0) { /* beginning of next page */
+			i++;
+			page = kaiocb->kaio_pages[i];
+		}
+	}
+	assert(!nbytes || pos == isize);
+	kaio_set_time(kaiocb, AIOTIME_COPY);
+
+error:
+
+	/* nbytes contains remainder unread, for eg., due to isize limit */
+	ret = kaio_io_complete(kaiocb, kaiocb->kaio_nbytes - nbytes, ret, 0);
+
+	return ret;
+}
+
+#if defined(CONFIG_RAW)
+static int
+do_raw_wait(kaiocb_t *kaiocb)
+{
+	int diag;
+
+	do {
+		diag = kaio_raw_wait(kaiocb->kaio_scp, &kaiocb->kaio_iobuf);
+	} while (diag == -EAGAIN);
+	diag = kaio_io_complete(kaiocb, diag, (diag > 0 ? 0 : diag), 0);
+	return diag;
+}
+
+static int
+do_kaio_raw_rw(int cmd, struct file *filp, struct kaiocb *kaiocb)
+{
+	int diag;
+	void *scp;
+
+	kaiocb->kaio_sem = MUTEX_LOCKED;
+	scp = kaio_raw_rw((cmd == AIOCMD_READ ? READ : WRITE), filp,
+			kaiocb->kaio_buf, kaiocb->kaio_nbytes,
+			&kaiocb->kaio_offset, &diag,
+			&kaiocb->kaio_sem, &kaiocb->kaio_iobuf);
+	kaiocb->kaio_scp = scp;
+	if (diag == 0) {
+		diag = kaio_enqueue(KAIO_INFO(current), kaiocb);
+		if (diag) {
+			do {
+				diag = kaio_raw_wait(kaiocb->kaio_scp,
+					&kaiocb->kaio_iobuf);
+			} while (diag == -EAGAIN);
+		}
+	}
+	return diag;
+}
+#endif /* CONFIG_RAW */
+
+kaiocb_t *susp_kaiocb;
+
+int
+kaio_suspend(aiocb_t *list[], int nent, struct timespec *utimeout, int nwait)
+{
+	int 			ret = 0, i, valid, pending;
+	struct timespec 	timeout;
+	aiocb_t			*uaiocb = 0;
+	kaio_key_t		kkey;
+	kaiocb_t 		*kaiocb, **kaiocbs, *kaiocb_array[256];
+	wait_queue_head_t	wq;
+	DECLARE_WAITQUEUE(wait, current);
+
+	if (utimeout && copy_from_user(&timeout, (struct timespec *)utimeout,
+				    sizeof(struct timespec)))
+	{
+		return -EFAULT;
+	}
+	/*
+	 * For more normal requests, allocate on stack.
+	 */
+	if (nent > 256) {
+		kaiocbs = kmalloc(sizeof(kaiocb_t *) * nent, GFP_KERNEL);
+		if (!kaiocbs)
+			return -EAGAIN;
+	} else
+		kaiocbs = kaiocb_array;
+
+	init_waitqueue_head(&wq);
+	add_wait_queue(&wq, &wait);
+
+	/*
+	 * Go through the list, if nwait done, return; otherwise
+	 * mark kaiocb so it will wake this thread when done.
+	 */
+	for (i = 0, valid = 0; nwait > 0 && i < nent; i++) {
+		if (get_user(uaiocb, &list[i])) {
+			ret = -EFAULT;
+			goto suspend_error;
+		}
+		if (!uaiocb)
+			continue;
+		/*
+		 * Check to see if done, otherwise put this
+		 * thread on the list of things to wake up.
+		 */
+		if (copy_from_user(&kkey, &uaiocb->aio_key,
+			sizeof(kaio_key_t)))
+		{
+			ret = -EFAULT;
+			goto suspend_error;
+		}
+		spin_lock(&kaio_hash_lock);
+		if ((kaiocb = kaio_find_hash(&kkey)) == 0) {
+			/* get_user can fault, better drop lock */
+			spin_unlock(&kaio_hash_lock);
+			if (get_user(ret, &uaiocb->aio_error)) {
+				ret = -EFAULT;
+				goto suspend_error;
+			}
+			/*
+			 * Assume that user gave a valid aiocb,
+			 * and that it is done.
+			 */
+			nwait--;
+		} else {
+			if (kaiocb->kaio_suspend_wait) {
+				/*
+				 * Don't allow two different processes to
+				 * wait on the same I/O request.
+				 */
+				ret = -EAGAIN;
+				spin_unlock(&kaio_hash_lock);
+				goto suspend_error;
+			}
+			kaiocb->kaio_suspend_wait = &wq;
+			kaiocbs[valid++] = kaiocb;
+			spin_unlock(&kaio_hash_lock);
+		}
+	}
+	if (!nwait && !valid)
+		return ret;
+	
+	susp_kaiocb = kaiocbs[0];
+
+	/*
+	 * Pending keeps track of how many I/Os  are being
+	 * wait upon. Valid is the index into kaiocbs where
+	 * the kaiocb of the last pending I/O is stored.
+	 * After sleeping & waking up on any one of the pending
+	 * I/O to complete, we scan the kaiocbs, garbage collect
+	 * all complete kaiocb & put a null entry in kaiocbs[].
+	 */
+	pending = valid;
+	lock_kernel();
+	while (nwait > 0 && pending) {
+		/*
+		 * See how many more done & free kaiocb's as necessary.
+		 */
+		for (i = 0; i < valid; i++) {
+			if (kaiocbs[i]->kaio_error != EINPROGRESS &&
+			     kaiocbs[i]->kaio_suspend_wait)
+			{
+				nwait--;
+				pending--;
+				kaiocbs[i]->kaio_suspend_wait = 0;
+			}
+			/* already done ?! */
+			if (!pending || nwait <= 0)
+				goto done_wait;
+		}
+		if (signal_pending(current)) {
+			ret = -ERESTARTSYS;
+			break;
+		}
+		current->state = TASK_INTERRUPTIBLE;
+		if (utimeout)
+			(void) schedule_timeout(timespec_to_jiffies(&timeout));
+		else
+			schedule();
+		current->state = TASK_RUNNING;
+		if (signal_pending(current)) {
+			ret = -ERESTARTSYS;
+			break;
+		}
+	}
+done_wait:
+	unlock_kernel();
+
+suspend_error:
+	spin_lock(&kaio_hash_lock);
+	/*
+	 * Unmark & free kaiocb's as necessary.
+	 */
+	for (i = 0; i < valid; i++) {
+		kaiocbs[i]->kaio_suspend_wait = 0;
+		if (kaiocbs[i]->kaio_error != EINPROGRESS) {
+			kaio_remove_hash(kaiocbs[i]);
+			fput(kaiocbs[i]->kaio_filp);
+			kfree(kaiocbs[i]);
+		}
+	}
+	spin_unlock(&kaio_hash_lock);
+	if (kaiocbs != kaiocb_array)
+		kfree(kaiocbs);
+	return(ret);
+}
+
+/*
+ * kaio_cancel: if given kkey, cancel a specific aio; otherwise
+ * 		cancel all aios for given fd. All successful canceled
+ *		aios will have -1 return & ECANCELED as aio_error.
+ *		kaio_io_complete handles normal signal delivery and
+ *		any listio that the aio may be part of.
+ *	kernel lock held on call.
+ */
+int
+kaio_cancel(int fd, kaio_key_t *kkey)
+{
+	kaiocb_t 	*kaiocb;
+	struct file 	*filp;
+	int		ret = 0;
+
+	if (!(filp = fget(fd)))
+		return -EBADF;
+
+	if (kkey) {
+		spin_lock(&kaio_hash_lock);
+		kaiocb = kaio_find_hash(kkey);
+		spin_unlock(&kaio_hash_lock);
+		if (kaiocb == 0) {
+			ret = -EINVAL;
+			goto cancel_out;
+		}
+
+		if (filp != kaiocb->kaio_filp) {
+			ret = -EBADF;
+			goto cancel_out;
+		}
+
+		if (!kaio_dequeue_kaiocb(KAIO_INFO(current), kaiocb)) {
+			ret = AIO_NOTCANCELED;
+			goto cancel_out;
+		}
+		ret = kaio_io_complete(kaiocb, -1, -ECANCELED, 1);
+		if (ret == 0)
+			ret = AIO_CANCELED;
+	} else {
+		/* cancel all associated with fd */
+		if ((kaiocb = kaio_dequeue_fd(KAIO_INFO(current), filp)) == 0) {
+			ret = AIO_ALLDONE;
+			goto cancel_out;
+		}
+		do {
+			/* XXX: what if error occurs */
+			kaio_io_complete(kaiocb, -1, -ECANCELED, 1);
+		} while ((kaiocb = kaio_dequeue_fd(KAIO_INFO(current), filp)) != 0);
+	}
+cancel_out:
+	fput(filp);
+	return ret;
+}
+
+void
+kaio_fork(struct task_struct *p, int clone_flags)
+{
+	/* Assumes kernel lock held */
+
+	/*
+	 * If the new thread is not sharing VM with parent,
+	 * then it can't share AIO state.
+	 */
+	if (clone_flags & CLONE_VM) {
+		KAIO_INFO(p) = KAIO_INFO(current);
+		KAIO_INFO(p)->kaioqref++;
+	} else
+		p->kaio_info = 0;
+}
+
+/*
+ * kaio_exit:
+ *	Called by generic exit when a thread has async I/O (kaio_info).
+ *	This routine does the following:
+ * 	(1) kills the slave threads, by signalling and
+ * 	     waiting for them to be reaped.
+ *	(2) Removes any queued aios (these may not have been intiated at all).
+ *	(3) Removes any remaining aios for this thread from the hash (these
+ *		have completed but not reaped through aio_suspend).
+ *	All actions protected by kernel lock.
+ */
+void
+kaio_exit(void)
+{
+	struct task_struct 	*p;
+	kaiocb_t 		**khpp;
+	kaio_queue_t		*kq;
+	struct list_head 	*tmp;
+	kaiocb_t 		*curr, *next;
+
+	lock_kernel();
+
+	/* kill all slaves if last clone */
+
+	kq = KAIO_INFO(current);
+	if (kq->kaioqref != 1) {
+		kq->kaioqref--;
+		unlock_kernel();
+		return;
+	}
+	/*
+	 * Last reference, kill all slaves & cleanup pending I/Os.
+	 */
+
+	for_each_task(p) {
+		/*
+		 * Look for slaves by name.
+		 */
+		if (p->mm == current->mm &&
+		     strncmp(&p->comm[0], KAIO_SLAVE_NAME,
+			     KAIO_SLAVE_NAMELEN) == 0)
+		{ 
+			send_sig(SIGKILL, p, 1);
+		}
+	}
+	/* kernel lock held for the test below */
+	while (kq->kaioq_nslaves != 0)
+		down(&kq->kaioq_exit_sem);
+	/*
+	 * Any queued aios can be simply "completed" & garbage collected. 
+	 */
+	for (tmp = kq->kaioq.prev; tmp != &kq->kaioq; tmp = kq->kaioq.prev) {
+		list_del(tmp);
+		kaio_io_complete(list_entry(tmp, struct kaiocb, kaio_ioq),
+					-1, -1, 1);
+	}
+	/*
+	 * Free up all hashed aios for this pid; these are already complete,
+	 * but pending aio_suspend.
+	 */
+	unlock_kernel();
+	spin_lock(&kaio_hash_lock);
+	for (khpp = &kaio_hash_table[0];
+		khpp < &kaio_hash_table[kaio_hash_table_size]; khpp++)
+	{
+
+		for  (curr = *khpp; curr; curr = next) {
+			next = curr->kaio_h_next;
+			if (curr->kaio_kq == kq) {
+				kaio_remove_hash(curr);
+				kfree(curr);
+			}
+		}
+	}
+	spin_unlock(&kaio_hash_lock);
+	kfree(kq);
+}
+
+#define KAIO_DO_FOP(op, k)  {						      \
+	int ret;							      \
+	(k)->kaio_filp->f_pos = (k)->kaio_offset;                             \
+	ret = (k)->kaio_filp->f_op-> ## op ((k)->kaio_filp,                   \
+		(k)->kaio_buf, (k)->kaio_nbytes, &(k)->kaio_offset);     \
+	kaio_io_complete((k), ret < 0 ? 0 : ret, ret > 0 ? 0 : ret, 0);	      \
+}
+
+static void
+kaio_set_slave_name(void)
+{
+	char newcomm[16];
+
+	sprintf(newcomm, "%s%05d", KAIO_SLAVE_NAME, current->p_pptr->pid);
+	strcpy(current->comm, newcomm);
+	current->mm->arg_start = current->mm->arg_end = 0;
+}
+
+/*
+ * kaio_slave:
+ * 	Entry point for slave threads that service queued I/O requests.
+ * For each task that employs async I/O there may be several slave threads.
+ * 
+ *	The main function of the slave threads is to wait & dequeue an I/O
+ * request and serice it. Requests may be split-phase with a seperate initiate
+ * and wait routine, or with a unified routine that does both.  Not all
+ * file systems support split-phase requests; only raw I/O directly support it.
+ * Additionally, for file systems that employ generic_read we can use
+ * the recoded kaio_generic_read & associated wait (kaio_slave_suspend).
+ *	
+ *	Slave threads are explicitly started by the library, but are killed
+ * when the "parent" task (that initiates I/O) exits (kaio_exit).
+ */
+void
+kaio_slave()
+{
+	kaiocb_t *kaiocb;
+	kaio_queue_t *kq;
+
+	if (current->p_pptr->pid == 1) { /* too late */
+		unlock_kernel();
+		return;
+	}
+
+	kaio_set_slave_name();
+	kq = KAIO_INFO(current->p_pptr); /* service parent's kaio queue */
+	KAIO_INFO(current) = 0; /* so kaio_exit does not trip */
+	kq->kaioqref--;
+	if (kq->kaioq_nslaves == -1)
+		kq->kaioq_nslaves = 1; /* initialize it */
+	else
+		kq->kaioq_nslaves++;
+	unlock_kernel();
+
+drain_io:
+	while ((kaiocb = kaio_dequeue(kq))) { /* blocks */
+#if defined(CONFIG_RAW)
+		if (kaio_is_raw(kaiocb->kaio_filp))
+			do_raw_wait(kaiocb);
+		else
+#endif
+		{
+			switch (kaiocb->kaio_cmd) {
+			case AIOCMD_GENERIC_READ:
+				kaio_slave_suspend(kaiocb);
+				break;
+			case AIOCMD_READ:
+				KAIO_DO_FOP(read, kaiocb);
+				break;
+			case AIOCMD_WRITE: {
+				KAIO_DO_FOP(write, kaiocb);
+				break;
+			 }
+			 default:
+				kaio_io_complete(kaiocb, 0, -EINVAL, 0);
+			}
+		}
+	}
+
+	spin_lock(&kq->kaioq_lock);
+	if (atomic_read(&kq->kaiosem.count) > 0) {
+		/*
+		 * There are queued I/Os; process them now.
+		 */
+		spin_unlock(&kq->kaioq_lock);
+		goto drain_io;
+	}
+	lock_kernel();
+	kq->kaioq_nslaves--;
+	spin_unlock(&kq->kaioq_lock);
+	up(&kq->kaioq_exit_sem);
+	unlock_kernel();
+}
+
+/*
+ * This is an asynchronous generic file read routine, and uses the
+ * inode->i_op->readpage() function for the actual low-level stuff.
+ */
+static inline int
+do_kaio_generic_read(struct file * filp, kaiocb_t *kaiocb)
+{
+	struct inode *inode = filp->f_dentry->d_inode;
+	size_t pos;
+	struct address_space *mapping = inode->i_mapping;
+	struct page **hash, *page, *page_cache;
+	int i, ret = 0, npages;
+
+	pos = kaiocb->kaio_offset;
+
+	if (pos >= inode->i_size) {
+		/*
+		 * No I/O possible; specified offset beyond file end
+		 */
+		return -EINVAL;
+	}
+
+	i = pos + kaiocb->kaio_nbytes - 1; /* last byte in I/O */
+	if (i > inode->i_size)
+		i = inode->i_size;
+	npages = pageof(i) - pageof(pos) + 1;
+	/*
+	 * Allocate space to hold page pointers.
+	 */
+	if (npages == 1)
+		kaiocb->kaio_pages = &kaiocb->kaio_page_inline;
+	else {
+		kaiocb->kaio_pages =
+			kmalloc(sizeof(struct page *) * npages, GFP_KERNEL);
+		if (!kaiocb->kaio_pages) 
+			return -EAGAIN;
+	}
+	page_cache = 0;
+	pos = pageof(pos) * PAGE_SIZE;
+	for (i = 0; i < npages; i++, pos += PAGE_SIZE) {
+
+retry:
+		/*
+		 * Try to find the data in the page cache..
+		 */
+		hash = page_hash(mapping, pageof(pos));
+		spin_lock(&pagecache_lock);
+		page = __find_page_nolock(mapping, pageof(pos), *hash);
+		if (page) {
+			get_page(page);
+			kaiocb->kaio_pages[i] = page;
+			spin_unlock(&pagecache_lock);
+			continue;
+		}
+
+		/*
+		 * Ok, it wasn't cached, so we need to create a new
+		 * page.. or reuse a page previously allocated,
+		 * and did not use in retry path below (found page in
+		 * hash after allocation).
+		 */
+		if (!page_cache) {
+			spin_unlock(&pagecache_lock);
+			page_cache = page_cache_alloc();
+			/*
+			 * That could have slept, so go around to the
+			 * very beginning..
+			 */
+			if (page_cache)
+				goto retry;
+			ret = -ENOMEM;
+			break;
+		}
+
+		/*
+		 * Ok, add the new page to the page cache.
+		 */
+		page = page_cache;
+		__add_to_page_cache(page, mapping, pageof(pos), hash);
+		spin_unlock(&pagecache_lock);
+		page_cache = 0;
+		kaiocb->kaio_pages[i] = page;
+
+		/*
+		 * Fire off the I/O; read page may also add page's
+		 * bufferheads to buffer cache.
+		 */
+		if ((ret = mapping->a_ops->readpage(filp, page)))
+			break;
+		run_task_queue(&tq_disk); /* XXX - improves perf. but why ??*/
+	}
+	if (!ret) {
+		kaiocb->kaio_npages = npages;
+		ret = kaio_enqueue(KAIO_INFO(current), kaiocb);
+	}
+	if (ret) { /* recheck return from kaio_enqueue */
+		for (i--; i >=0; i--) {
+			assert(kaiocb->kaio_pages[i]);
+			page_cache_free(kaiocb->kaio_pages[i]);
+		}
+		if (kaiocb->kaio_pages != &kaiocb->kaio_page_inline)
+			kfree(kaiocb->kaio_pages);
+	}
+	if (page_cache)
+		page_cache_free(page_cache);
+	UPDATE_ATIME(inode);
+	return ret;
+}
+
+/*
+ * Initiate buffered async I/O (e.g. file systems).
+ *
+ * Kaiocb describing the I/O is queued as part of a per-task queue (KAIO_INFO).
+ * I/O is actually initiated only in the case of generic_read. For all
+ * other cases, slave threads initiate and wait for queued requests.
+ *
+ */
+static inline int
+do_kaio_fs_rw(int cmd, struct file *filp, struct kaiocb *kaiocb)
+{
+	/*
+	 * if read request & the fop operation is generic,
+	 * use our own parallel async version; for all other
+	 * reads & for writes, asynchronize by queuing f_op to be
+	 * executed directly by slave.
+	 */
+	if (cmd == AIOCMD_READ && filp->f_op->read == generic_file_read) {
+		kaiocb->kaio_cmd = AIOCMD_GENERIC_READ;
+		return(do_kaio_generic_read(filp, kaiocb));
+	} else {
+		return(kaio_enqueue(KAIO_INFO(current), kaiocb));
+	}
+}
+
+/*
+ * kaio_rw: Initiate async I/O read or write.
+ *
+ *	cmd 	= AIOCMD_{READ,WRITE}; ignored for listio.
+ *	uaiocb 	= user space address of aiocb describing request.
+ *	liocb 	= if not null then part of a list io. 
+ *
+ * Kernel lock held across this function. If I/O successfully queued,
+ * kaiocb holds reference to file (with fget, correspoding fput will be done
+ * at io completion/cancellation/exit). Kaiocb also holds references to the
+ * requesting task, the uaiocb (for transmitting back return/error values),
+ * and liocb, if any. A unique id, representing the  kaiocb in a hash table,
+ * is "returned" back to user on successful queueing.
+ *
+ */
+int
+kaio_rw(int cmd, aiocb_t *uaiocb, liocb_t *liocb)
+{
+	struct file 	*filp = NULL;
+	aiocb_t		aiocb;
+	int 		ret = 0;
+	kaiocb_t 	*kaiocb = NULL;
+
+
+	if (copy_from_user(&aiocb, uaiocb, sizeof(aiocb_t)))
+		return -EFAULT;
+
+	if (kaio_nr_hashed > CONFIG_AIO_MAX) {
+		ret = -EAGAIN;
+		goto rw_error;
+	}
+	if (liocb) {
+		switch (aiocb.aio_lio_opcode) {
+		case LIO_READ:
+			cmd = AIOCMD_READ;
+			break;
+		case LIO_WRITE:
+			cmd = AIOCMD_WRITE;
+			break;
+		case LIO_NOP:
+			/*
+			 * No easy way to return to caller an
+			 * indication to ignore; so, alternately,
+			 * mark as complete.
+			 */
+			if (!lio_complete(liocb, 1, 0))
+				unlock_kernel(); /* not really sleeping */
+			return 0;
+		default:
+			ret =  -EINVAL;
+			goto rw_error;
+		}
+	}
+	if (!(filp = fget(aiocb.aio_fildes))) {
+		ret = -EBADF;
+		goto rw_error;
+	}
+	if (!(filp->f_mode &
+		(cmd == AIOCMD_READ ? FMODE_READ: FMODE_WRITE)))
+	{
+		ret = -EBADF;
+		goto rw_error;
+	}
+	if (aiocb.aio_nbytes) {
+
+		kaiocb = kmalloc(sizeof(kaiocb_t), GFP_KERNEL);
+		if (!kaiocb) {
+			ret = -EAGAIN;
+			goto rw_error;
+		}
+		kaiocb->kaio_filp = filp;
+		kaiocb->kaio_buf = aiocb.aio_buf;
+		kaiocb->kaio_nbytes = aiocb.aio_nbytes;
+		kaiocb->kaio_offset = aiocb.aio_offset;
+		kaiocb->kaio_sigevent = aiocb.aio_sigevent;
+		kaiocb->kaio_task = current;
+		kaiocb->kaio_kq = KAIO_INFO(current);
+		kaiocb->kaio_uaiocb = uaiocb;
+		kaiocb->kaio_cmd = cmd;
+		kaiocb->kaio_liocb = liocb;
+		kaiocb->kaio_pages = 0;
+		if (cmd == AIOCMD_READ &&
+			!access_ok(VERIFY_WRITE, aiocb.aio_buf, aiocb.aio_nbytes))
+		{
+			ret = -EFAULT;
+			goto rw_error;
+		}
+		ret = locks_verify_area(cmd == AIOCMD_READ ?
+					FLOCK_VERIFY_READ : FLOCK_VERIFY_WRITE,
+					filp->f_dentry->d_inode, filp,
+					aiocb.aio_offset, aiocb.aio_nbytes);
+		if (ret)
+			goto rw_error;
+#if defined(CONFIG_RAW)
+		if (kaio_is_raw(filp))
+			ret = do_kaio_raw_rw(cmd, filp, kaiocb);
+		else
+#endif
+			ret = do_kaio_fs_rw(cmd, filp, kaiocb);
+	}
+	if (!ret) {
+		kaio_key_t kkey;
+
+		kkey.kaio_kid = kaiocb->kaio_id;
+		kkey.kaio_kq = current->kaio_info;
+
+		if (copy_to_user(&uaiocb->aio_key, &kkey, sizeof(kkey)))
+			ret = -EFAULT;
+	}
+rw_error:
+	if (ret) {
+		if (filp)
+			fput(filp);
+		if (kaiocb)
+			kfree(kaiocb);
+	}
+	if (liocb && ret) {
+		/*
+		 * Have to set individual errors for list io
+		 */
+		 (void) put_user(-ret, &uaiocb->aio_error);
+	}
+	return ret;
+}
+
+
+static void
+kaio_set_list_errors(aiocb_t *list[], int nent, int error)
+{
+	int	i;
+	aiocb_t *uaiocbp;
+
+	/*
+	 * Have to set individual errors for each aiocb.
+	 */
+	for (i = 0; i < nent; i++)
+		if (get_user(uaiocbp, &list[i]) == 0 && uaiocbp)
+			 (void) put_user(-error, &uaiocbp->aio_error);
+}
+
+int
+kaio_list_io(int mode, struct aiocb *list[], int nent, sigevent_t *sig)
+{
+	liocb_t *liocb;
+	aiocb_t *aiocbp;
+	int	i = 0, ret = 0, ignore = 0, error = 0;
+
+	liocb = kmalloc(sizeof(liocb_t), GFP_KERNEL);
+	if (!liocb) {
+		ret = -EAGAIN;
+		kaio_set_list_errors(list, nent, ret);
+		return ret;
+	}
+
+	liocb->lio_lock = SPIN_LOCK_UNLOCKED;
+	liocb->lio_nent = nent;
+	liocb->lio_done = 0;
+	liocb->lio_mode = mode;
+	init_waitqueue_head(&liocb->lio_wait);
+	liocb->lio_task = current;
+	if (sig) {
+		if (copy_from_user(&liocb->lio_sigevent,
+			sig, sizeof(sigevent_t)))
+		{
+			ret =  -EFAULT;
+			kaio_set_list_errors(list, nent, ret);
+			error = nent; /* all are in error */
+			goto lio_error;
+		}
+	} else
+		liocb->lio_sigevent.sigev_notify = SIGEV_NONE;
+	for ( ; i < nent; i++) {
+		if (get_user(aiocbp, &list[i])) {
+			ret = -EFAULT;
+			error++;
+			continue;
+		}
+		if (!aiocbp) {
+			ignore++;
+			continue;
+		}
+		/* 0 == cmd for kaio_rw derived from aiocbp */
+		if ((ret = kaio_rw(0, aiocbp, liocb)))
+			error++;
+	}
+	if (!error) {
+		if (!lio_complete(liocb, ignore, 0)) {
+			if (mode == LIO_WAIT)
+				interruptible_sleep_on(&liocb->lio_wait);
+			unlock_kernel();
+		}
+		if (mode == LIO_WAIT)
+			kfree(liocb);
+		return 0;
+	}
+lio_error:
+	if (!lio_complete(liocb, (ignore + error), 0)) {
+		/* pending I/O's; but we are not interested */
+		liocb->lio_mode = LIO_NOWAIT;
+		liocb->lio_sigevent.sigev_notify = SIGEV_NONE;
+		unlock_kernel();
+	} else if (mode != LIO_NOWAIT) {
+		/* if no wait then lio_complete frees liocb */
+		kfree(liocb);
+	}
+	return ret;
+}
+#endif /* CONFIG_AIO */
 
 /*
  * This is the "read()" routine for all filesystems
